from leanai.data.datasets.fashion_mnist_dataset import FashionMNISTDataset
from leanai.data.datasets.casia_faces_dataset import CasiaFacesDataset
from leanai.data.datasets.coco_dataset import COCODataset
from leanai.data.datasets.images_in_classfolders_dataset import ImagesInClassfoldersDataset
